# This test creates a video to showcase the pre-training (last layer) and training of the DBN
# As it is very slow it is disabled by default

output.folder <- "video"
maxiters.pretrain <- 1e6
maxiters.train <- 10000
run.training <- run.images <- TRUE
highlight.digits = c(72, 3, 83, 91, 6688, 7860, 92, 1, 180, 13)

library(DeepLearning)
library(mnist)
data(mnist)

# Define the network
dbn <- DeepBeliefNet(Layers(c(784, 1000, 500, 250, 2), input = "continuous", output = "gaussian"), initialize = "0")

format.timediff <- function(start.time) {
    diff = as.numeric(difftime(Sys.time(), start.time, units="mins"))
    hr <- diff%/%60
    min <- floor(diff - hr * 60)
    sec <- round(diff%%1 * 60,digits=2)
    return(paste(hr,min,sec,sep=':'))
}

if (run.training) {
	sprintf.fmt.iter <- sprintf("%%0%dd", nchar(sprintf("%d", maxiters.pretrain)))
	mnist.data.layer <- mnist
	for (i in 1:3) {
		rbm <- dbn[[i]]
		diag <- list(rate = "accelerate", data = NULL, f = function(rbm, batch, data, iter, batchsize, maxiters, layer) {
			print(sprintf("%s[%s/%s] in %s", layer, iter, maxiters, format.timediff(start.time)))
		})
		save(rbm, file = file.path(output.folder, sprintf("rbm-%s-%s.RData", i, "initial")))
		start.time <- Sys.time()
		rbm <- pretrain(rbm, mnist.data.layer$train$x,  penalization = "l2", lambda=0.0002, momentum = c(0.5, 0.9),
							 epsilon=c(.1, .1, .1, .001)[i], batchsize = 100, maxiters=maxiters.pretrain,
							 continue.function = continue.function.always, diag = diag)
		mnist.data.layer$train$x <- predict(rbm, mnist.data.layer$train$x)
		mnist.data.layer$test$x <- predict(rbm, mnist.data.layer$test$x)
		save(rbm, file = file.path(output.folder, sprintf("rbm-%s-%s.RData", i, "final")))
		dbn[[i]] <- rbm
	}
	
	rbm <- dbn[[4]]
	diag <- list(rate = "accelerate", data = NULL, f = function(rbm, batch, data, iter, batchsize, maxiters, layer) {
		save(rbm, file = file.path(output.folder, sprintf("rbm-4-%s.RData", sprintf(sprintf.fmt.iter, iter))))
		print(sprintf("%s[%s/%s] in %s", layer, iter, maxiters, format.timediff(start.time)))
	})
	save(rbm, file = file.path(output.folder, sprintf("rbm-%s-%s.RData", 4, "initial")))
	start.time <- Sys.time()
	rbm <- pretrain(rbm, mnist.data.layer$train$x,  penalization = "l2", lambda=0.0002,
						 epsilon=.001, batchsize = 100, maxiters=maxiters.pretrain,
						 continue.function = continue.function.always, diag = diag)
	save(rbm, file = file.path(output.folder, sprintf("rbm-4-%s.RData", "final")))
	dbn[[4]] <- rbm
	
	# Fine-tune
	sprintf.fmt.iter <- sprintf("%%0%dd", nchar(sprintf("%d", maxiters.train)))
	diag <- list(rate = "each", data = NULL, f = function(dbn, batch, data, iter, batchsize, maxiters) {
		save(dbn, file = file.path(output.folder, sprintf("dbn-finetune-%s.RData", sprintf(sprintf.fmt.iter, iter))))
		print(sprintf("[%s/%s] in %s", iter, maxiters, format.timediff(start.time)))
	})
	save(dbn, file = file.path(output.folder, sprintf("dbn-finetune-%s.RData", "initial")))
	start.time <- Sys.time()
	dbn <- train(unroll(dbn), mnist$train$x, batchsize = 100, maxiters=maxiters.train,
					continue.function = continue.function.always, diag = diag)
	save(dbn, file = file.path(output.folder, sprintf("dbn-finetune-%s.RData", "final")))
	
}

if (run.images) {
	for (i in 1:3) {
		load(file.path(output.folder, sprintf("rbm-%d-final.RData", i)))
		dbn[[i]] <- rbm
	}
	
	for (file in list.files(output.folder, pattern = "rbm-4-.+\\.RData", full.names = TRUE)) {
		print(file)
		load(file)
		dbn[[4]] <- rbm
		iter <- stringr::str_match(file, "rbm-4-(.+)\\.RData")[,2]
		#png(sub(".RData", ".png", file), width = 1920, height = 1080) # full hd output
		png(sub(".RData", ".png", file), width = 1280, height = 720) # hd output
		predictions <- predict(dbn, mnist$test$x)
		reconstructions <- reconstruct(dbn, mnist$test$x)
		iteration.error <- errorSum(dbn, mnist$test$x) / nrow(mnist$test$x)
		plot.mnist(model = dbn, x = mnist$test$x, label = mnist$test$y+1, predictions = predictions, reconstructions = reconstructions,
				   ncol = 16, highlight.digits = highlight.digits,
				   xlim = c(-12.625948, 8.329168), ylim = c(-10.50657, 13.12654))
		par(family="mono")
		legend("bottomleft", legend = sprintf("Mean error = %.3f", iteration.error), bty="n", cex=3)
		legend("bottomright", legend = sprintf("Iteration = %s", iter), bty="n", cex=3)
		dev.off()
	}

	# Fine-tuning
	for (file in list.files(output.folder, pattern = "dbn-finetune-.+\\.RData", full.names = TRUE)) {
		print(file)
		load(file)
		iter <- stringr::str_match(file, "dbn-finetune-(.+)\\.RData")[,2]
		#png(sub(".RData", ".png", file), width = 1920, height = 1080) # full hd output
		png(sub(".RData", ".png", file), width = 1280, height = 720) # hd output
		predictions <- predict(dbn, mnist$test$x)
		reconstructions <- reconstruct(dbn, mnist$test$x)
		iteration.error <- errorSum(dbn, mnist$test$x) / nrow(mnist$test$x)
		plot.mnist(model = dbn, x = mnist$test$x, label = mnist$test$y+1, predictions = predictions, reconstructions = reconstructions,
				   ncol = 16, highlight.digits = highlight.digits,
				   xlim = c(-22.81098,  27.94829), ylim = c(-17.49874,  33.34688))
		par(family="mono")
		legend("bottomleft", legend = sprintf("Mean error = %.3f", iteration.error), bty="n", cex=3)
		legend("bottomright", legend = sprintf("Iteration = %s", iter), bty="n", cex=3)
		dev.off()
	}
}
